@@ -17,7 +17,7 @@ use crate::ll::Request as _;
 #[cfg(feature = "abi-7-21")]
 use crate::reply::ReplyDirectoryPlus;
 use crate::reply::{Reply, ReplyDirectory, ReplySender};
-use crate::session::Session;
+use crate::session::{Session, SessionACL};
 use crate::Filesystem;
 use crate::{ll, KernelConfig};
 
@@ -34,7 +34,7 @@ pub struct Request<'a> {
 
 impl<'a> Request<'a> {
     /// Create a new request from the given data
-    pub fn new(ch: ChannelSender, data: &'a [u8]) -> Option<Request<'a>> {
+    pub(crate) fn new(ch: ChannelSender, data: &'a [u8]) -> Option<Request<'a>> {
         let request = match ll::AnyRequest::try_from(data) {
             Ok(request) => request,
             Err(err) => {
@@ -49,7 +49,7 @@ impl<'a> Request<'a> {
     /// Dispatch request to the given filesystem.
     /// This calls the appropriate filesystem operation method for the
     /// request and sends back the returned reply to the kernel
-    pub fn dispatch<FS: Filesystem>(&self, se: &mut Session<FS>) {
+    pub(crate) fn dispatch<FS: Filesystem>(&self, se: &mut Session<FS>) {
         debug!("{}", self.request);
         let unique = self.request.unique();
 
@@ -59,15 +59,61 @@ impl<'a> Request<'a> {
             Err(errno) => self.request.reply_err(errno),
         }
         .with_iovec(unique, |iov| self.ch.send(iov));
+
         if let Err(err) = res {
             warn!("Request {:?}: Failed to send reply: {}", unique, err)
         }
     }
+
     fn dispatch_req<FS: Filesystem>(
         &self,
         se: &mut Session<FS>,
     ) -> Result<Option<Response>, Errno> {
         let op = self.request.operation().map_err(|_| Errno::ENOSYS)?;
+        // Implement allow_root & access check for auto_unmount
+        if (se.allowed == SessionACL::RootAndOwner
+            && self.request.uid() != se.session_owner
+            && self.request.uid() != 0)
+            || (se.allowed == SessionACL::Owner && self.request.uid() != se.session_owner)
+        {
+            #[cfg(feature = "abi-7-21")]
+            {
+                match op {
+                    // Only allow operations that the kernel may issue without a uid set
+                    ll::Operation::Init(_)
+                    | ll::Operation::Destroy(_)
+                    | ll::Operation::Read(_)
+                    | ll::Operation::ReadDir(_)
+                    | ll::Operation::ReadDirPlus(_)
+                    | ll::Operation::Write(_)
+                    | ll::Operation::FSync(_)
+                    | ll::Operation::FSyncDir(_)
+                    | ll::Operation::Release(_)
+                    | ll::Operation::ReleaseDir(_) => {}
+                    _ => {
+                        return Err(Errno::EACCES);
+                    }
+                }
+            }
+            #[cfg(not(feature = "abi-7-21"))]
+            {
+                match op {
+                    // Only allow operations that the kernel may issue without a uid set
+                    ll::Operation::Init(_)
+                    | ll::Operation::Destroy(_)
+                    | ll::Operation::Read(_)
+                    | ll::Operation::ReadDir(_)
+                    | ll::Operation::Write(_)
+                    | ll::Operation::FSync(_)
+                    | ll::Operation::FSyncDir(_)
+                    | ll::Operation::Release(_)
+                    | ll::Operation::ReleaseDir(_) => {}
+                    _ => {
+                        return Err(Errno::EACCES);
+                    }
+                }
+            }
+        }
         match op {
             // Filesystem initialization
             ll::Operation::Init(x) => {
@@ -108,7 +154,7 @@ impl<'a> Request<'a> {
             }
             // Filesystem destroyed
             ll::Operation::Destroy(x) => {
-                se.filesystem.destroy(self);
+                se.filesystem.destroy();
                 se.destroyed = true;
                 return Ok(Some(x.reply()));
             }
